##########################################################################
## Sample from the posterior distribution of a hierarchical ordered
## probit model via data augmentation.
##
## This software is distributed under the terms of the GNU GENERAL
## PUBLIC LICENSE Version 2, June 1991.  See the package LICENSE
## file for more information.
##
## PDB 05/26/2009
##
## Copyright (C) 2009-present Paul D. Baines and Kevin M. Quinn
##########################################################################


"MCMChieroprobit" <-
  function(formula, fullW, Wtilde, m,
           data = parent.frame(), burnin = 1000, mcmc = 10000,
           thin = 1, tune = NA, tdf = 1, verbose = 0, seed = NA, beta.start = NA,
           b0 = 0, B0 = 0, a0 = 0, A0 = 0, r0=0, R0=0, 
           printevery = 5000, ...) 
{
    ## checks
    check.offset(list(...))
    check.mcmc.parameters(burnin, mcmc, thin)
    
    ## seeds
    seeds <- form.seeds(seed) 
    lecuyer <- seeds[[1]]
    seed.array <- seeds[[2]]
    lecuyer.stream <- seeds[[3]]
    
    ## extract X, Y, W and variable names from the model formula and frame 
	call <- match.call() 	  
	mt <- terms(formula, data=data)
	if(missing(data)) data <- sys.frame(sys.parent())
    mf <- match.call(expand.dots = FALSE)
    mf$burnin <- mf$mcmc <- mf$b0 <- mf$B0 <- mf$a0 <- mf$A0 <- NULL
    mf$thin <- mf$... <- mf$tune <- mf$tdf <- mf$verbose <- mf$seed <- NULL
    mf$m <- mf$r0 <- mf$R0 <- NULL
    mf$beta.start <- mf$mcmc.method <- NULL
    mf$drop.unused.levels <- TRUE
    mf[[1]] <- as.name("model.frame")
    mf <- eval(mf, sys.frame(sys.parent()))
    vars <- as.character(attr(mt, "variables"))[-1] # y varname and x varnames
    
    ## null model support
    X <- if (!is.empty.model(mt)) model.matrix(mt, mf, contrasts)# else NULL
    X.names <- dimnames(X)[[2]]
    Y    <- model.response(mf, "numeric")
    Y    <- factor(Y, ordered=TRUE)
    ncat <- nlevels(Y)             # number of categories of y
    cat  <- levels(Y)              # values of categories of y
    N <- nrow(X)	               # number of observations
    K <- ncol(X)	               # number of covariates

                       # Vector containing the cluster sizes
                       # NOTE: Currently, rows 1:m[1] must correspond to cluster 1,
                       # rows[(m[1]+1):(m[1]+m[2])] to cluster 2 etc.
    m <- matrix(as.integer(m),ncol=1)
    if (any(m) <= 0){
      cat("All cluster sizes must be positive.\n")
      stop("Please respecify and call MCMChieroprobit() again.\n")      
    }
    if (sum(m) != N){
      cat("Sum of cluster sizes must equal number of rows of X.\n")
      stop("Please respecify and call MCMChieroprobit() again.\n")      
    }
    n <- nrow(m)     # Number of clusters

    if (length(Y) != N){
      cat("X and Y do not have same number of rows.\n")
      stop("Please respecify and call MCMChieroprobit() again.\n")      
    }
    
    ## convert data to matrices to be passed
    Y <- as.matrix(as.integer(Y))
    X <- as.matrix(X)
    
    ## check tuning parameter
    if (is.na(tune)){
      tune <- 0.05/ncat
    }
    
    xint <- match("(Intercept)", colnames(X), nomatch=0)
    if (xint > 0){
      new.X <- X[, -xint, drop=FALSE]
    }
    else {
      new.X <- X
      if (length(unique(new.X[,1])) && unique(new.X[,1])==1){
        new.X <- new.X[,-1,drop=FALSE]
      } else {
        warning("An intercept is needed and assumed in MCMChieroprobit()\n.")
      }
    }

    ## starting values for beta error checking
    if (is.na(beta.start)){
      if (ncol(new.X) == 0){
        polr.out <- polr(ordered(Y)~1)
      }
      else {
        polr.out <- polr(ordered(Y)~new.X)
      }
      if (verbose){
        cat("Output of polr:\n")
        print(polr.out)
      }    
      beta.start <- matrix(0, K, 1)
      beta.start[1] <- -.588 * polr.out$zeta[1]
      if( ncol(new.X) > 0){
        beta.start[2:K] <- .588 * coef(polr.out)
      }
    }
    else if(is.null(dim(beta.start))) {
      beta.start <- beta.start * matrix(1,K,1)  
    }
    else if((dim(beta.start)[1] != K) || (dim(beta.start)[2] != 1)) {
      cat("Starting value for beta not conformable.\n")
      stop("Please respecify and call MCMChieroprobit() again.\n")
    }

        ## formatting W
	fullW <- as.matrix(fullW)
	dim_xi <- ncol(fullW)  # Dimension of the random effects
	if(dim(fullW)[1] != N) {
		cat("Non-conformable size of fullW, must have N rows.\n")
		stop("Please respecify and call MCMChieroprobit() again.\n") 
	}   

        ## formatting Wtilde
	Wtilde <- as.matrix(Wtilde)
	if((dim(Wtilde)[1] != N) || (dim(Wtilde)[2] != n*dim_xi)) {
		cat("Non-conformable size of Wtilde, must be N x (n*dim_xi).\n")
		stop("Please respecify and call MCMChieroprobit() again.\n") 
	}   

    
	## prior for beta error checking
	if(is.null(dim(b0))) {
		b0 <- b0 * matrix(1,K,1)  
	}
	if((dim(b0)[1] != K) || (dim(b0)[2] != 1)) {
		cat("N(b0,B0) prior b0 not conformable.\n")
		stop("Please respecify and call MCMChieroprobit() again.\n") 
	}   
	if(is.null(dim(B0))) {
		B0 <- B0 * diag(K)    
	}
	if((dim(B0)[1] != K) || (dim(B0)[2] != K)) {
		cat("N(b0,B0) prior B0 not conformable.\n")
		stop("Please respecify and call MCMChieroprobit() again.\n")
	}

	## prior for alpha error checking
	if(is.null(dim(a0))) {
		a0 <- a0 * matrix(1, ncat-1, 1)  
	}
	if((dim(a0)[1] != ncat-1) || (dim(a0)[2] != 1)) {
		cat("N(a0,A0) prior a0 not conformable.\n")
		stop("Please respecify and call MCMChieroprobit() again.\n") 
	}   
	if(is.null(dim(A0))) {
		A0 <- A0 + diag(ncat - 1)    
	}
	if((dim(A0)[1] != ncat - 1) || (dim(A0)[2] != ncat - 1)) {
		cat("N(a0, A0) prior A0 not conformable.\n")
		stop("Please respecify and call MCMChieroprobit() again.\n")
	}
	## prior for Sigma_xi checking
	if (!is.null(dim(r0)) || r0<0.0){
		cat("r0 must be non-negative.\n")
		stop("Please respecify and call MCMChieroprobit() again.\n")
	}
	if(is.null(dim(R0))) {
		R0 <- R0 + diag(dim_xi)    
	}
	if((dim(R0)[1] != dim_xi) || (dim(R0)[2] != dim_xi)) {
		cat("Inv-Wishart(r0, R0) prior R0 not conformable.\n")
		stop("Please respecify and call MCMChieroprobit() again.\n")
	}
		
	## form gamma starting values (note: not changeable)
	gamma <- matrix(NA,ncat+1,1)
	gamma[1] <- -300
        gamma[2] <- 0
        if (ncat >2){
	  gamma[3:ncat] <- (polr.out$zeta[2:(ncat-1)] - polr.out$zeta[1])*.588
        }
        gamma[ncat+1] <- 300
	if (verbose){
		cat("Starting state for gamma:\n")
		print(gamma)
	}
		
	## posterior sample
	sample <- matrix(data=0, mcmc/thin, K + ncat + 1 + (dim_xi*(dim_xi+1)/2))

	## call C++ code to draw sample
	nY <- as.matrix(as.numeric(Y))
		
	## form the tuning parameter
	tune <- vector.tune(tune, ncat-1)
	posterior <- NULL

	xi_tilde <- as.matrix(as.numeric(rep(0,ncol(Wtilde))))
	U        <- as.matrix(as.numeric(rnorm(N)))
	printevery <- as.integer(printevery)

	if (verbose){
		cat("Calling C code...\n")
		cat(paste("nrow(nY)       = ",nrow(nY),      ", ncol(nY)       = ",ncol(nY),"\n",sep=""))
		cat(paste("nrow(X)        = ",nrow(X),       ", ncol(X)        = ",ncol(X),"\n",sep=""))
		cat(paste("nrow(fullW)        = ",nrow(fullW),       ", ncol(fullW)        = ",ncol(fullW),"\n",sep=""))
		cat(paste("nrow(Wtilde)   = ",nrow(Wtilde),  ", ncol(Wtilde)   = ",ncol(Wtilde),"\n",sep=""))
		cat(paste("nrow(xi_tilde) = ",nrow(xi_tilde),", ncol(xi_tilde) = ",ncol(xi_tilde),"\n",sep=""))
	}
	
	auto.Scythe.call(output.object="posterior", cc.fun.name="MCMChieroprobit",
                     sample.nonconst=sample, Y=as.integer(Y), nY=nY, X=X, 
                     U=U, m=m, fullW=fullW, tildeW=Wtilde, xi_tilde=xi_tilde,
                     burnin=as.integer(burnin),
                     mcmc=as.integer(mcmc), thin=as.integer(thin),
                     tune=tune, tdf=as.double(tdf), 
		     lecuyer=as.integer(lecuyer),
                     seedarray=as.integer(seed.array),
                     lecuyerstream=as.integer(lecuyer.stream),
                     verbose=as.integer(verbose), beta=beta.start,
                     gamma=gamma, b0=b0, B0=B0, a0=a0, A0=A0, r0=r0, R0=R0,
                     printevery=printevery)

	if (verbose){
		cat("Returned to R code...\n")
	}
		
    
	## put together matrix and build MCMC object to return
	sample <- matrix(posterior$sampledata, posterior$samplerow,
                     posterior$samplecol, byrow=FALSE)
	sample <- sample[, c(1:K, (K+3):(K+ncat), (K+ncat+2):ncol(sample))]

	output <- mcmc(data=sample, start=burnin+1, end=burnin+mcmc, thin=thin)
	xnames <- c(X.names, paste("gamma", 2:(ncat-1), sep=""))
	for (i in 1:dim_xi){
		for (j in i:dim_xi){
			xnames <- c(xnames,paste("Sigma_xi_",i,"_",j,"",sep=""))
		}
	}
	varnames(output) <- xnames
	attr(output, "title") <- "MCMChieroprobit Posterior Sample"
        
	return(output)
}
