//////////////////////////////////////////////////////////////////////////
// MCMChieroprobit.cc is C++ code to estimate a hierarchical 
//   ordinalprobit regression model with a multivariate normal prior
//
// Paul D. Baines
// Dept. of Statistics
// Harvard University
// baines@stat.harvard.edu
//
// Kevin M. Quinn
// Dept. of Government
// Harvard University
// kevin_quinn@harvard.edu
// 
// This software is distributed under the terms of the GNU GENERAL
// PUBLIC LICENSE Version 2, June 1991.  See the package LICENSE
// file for more information.
//
// Created 05/18/09 PDB
//
// Copyright (C) 2009-present Paul D. Baines and Kevin M. Quinn.
//////////////////////////////////////////////////////////////////////////


#ifndef MCMCHIEROPROBIT_CC
#define MCMCHIEROPROBIT_CC

#include "MCMCrng.h"
#include "MCMCfcds.h"
#include "matrix.h"
#include "distributions.h"
#include "stat.h"
#include "la.h"
#include "ide.h"
#include "smath.h"
#include "rng.h"
#include "mersenne.h"
#include "lecuyer.h"
#include "optimize.h"

#include <R.h>           // needed to use Rprintf()
#include <R_ext/Utils.h> // needed to allow user interrupts

using namespace std;
using namespace scythe;

static inline double hier_lnmulttdens(const Matrix<>& theta, 
                                      const Matrix<>& mu,
                                      const Matrix<>& C,
                                      const double df){
	
	const int d = theta.size();
	//const Matrix<> z = t(theta - mu) * C; 
	// C is now C' if VC mat is C C'
	const Matrix<> z = C * (theta - mu);
	double zsumsq = 0;
	for (int i=0; i<d; ++i){
		zsumsq += std::pow(z[i], 2);
	}
	
	return ( (-(df + d)/2) * std::log(1 + (1/df) * zsumsq)  );
}

void print_matrix(const Matrix<>& mat){
	for (unsigned int i=0; i<mat.rows(); i++){
		for (unsigned int j=0; j<mat.cols(); j++){
			Rprintf("%g\t",mat(i,j));
		}
		Rprintf("\n");
	}
	Rprintf("\n");
	return;
}

// function that transforms alpha to gamma 
Matrix<> hier_gamma2alpha(const Matrix<>& gamma){
	const int m = gamma.rows() - 2;
	Matrix<> alpha(m, 1);
    alpha[0] = std::log(gamma[1]);
    for (int j=1; j< m ; ++j){
        alpha[j] = std::log(gamma[j+1] - gamma[j]);
    }
    return alpha;
}

// function that transforms alpha to gamma 
Matrix<> hier_alpha2gamma(const Matrix<>& alpha){
	const int m = alpha.rows();
	Matrix<> gamma(m+2, 1);
		gamma[0] = -300;
		gamma[m+1] = 300;
	for (int j=1; j<m+1 ; ++j){
		double gamma_sum = 0.0;
		for(int i=0; i<j; ++i){
			gamma_sum += exp(alpha[i]);
		}
		gamma[j] = gamma_sum;
	}
	return gamma;
}

double hier_lndmvn_jhp(const Matrix<double> &x, const Matrix<double> &mu,
				const Matrix<double> &Sigma)
{
    int k = Sigma.cols();
    double first = (-k/2.0) * ::log(2 * M_PI) -0.5 * ::log(det(Sigma));
    Matrix< > second = ::t(x-mu)*invpd(Sigma)*(x-mu);    
    return (first - 0.5*second[0]);
}



// orpobit_logpost
static inline double hieroprobit_logpost(const Matrix<double>& nY, const Matrix<double>& X, 
							  const Matrix<double>& alpha,
							  const Matrix<double>& alpha_prior_mean, 
							  const Matrix<double>& alpha_prior_var,
							  const Matrix<double>& beta){
	
	//  likelihood
	double loglike = 0.0;
	const int n = nY.rows();
	const int ncat = alpha.rows() + 1;

	// the linear predictor
	Matrix<> mu = X*beta;
	Matrix<> gamma = hier_alpha2gamma(alpha);
	
	// compute prob
	Matrix<> cat_prob(n, ncat-1);
	//cat_prob: CATegory specific PROBability
	//the first col of cat_prob = pnorm(gamma_1 - mu)
	//thus, the col number is ncat - 1
	Matrix<> prob(n, ncat);
	for (int j=0; j< ncat-1; ++j){
		for (int i=0; i<n ; ++i){
			cat_prob(i, j) = pnorm(gamma[j+1] - mu[i], 0.0, 1.0);
		}
	}
	prob(_, ncat-1)  = 1 - cat_prob(_, ncat-2);    // top category
	prob(_, 0)       = cat_prob(_, 0);             // bottom category
	for (int j=1; j<(ncat-1); ++j){               // middle categories are actually differences of cumulatives
		prob(_, j) = cat_prob(_,j) - cat_prob(_, j-1);
	}
	
	// the loglikelihood

	for (int i=0; i<n; ++i){
		int ind = (int) nY[i];
		loglike += std::log(prob(i,ind-1));
	}
	
	// prior
	double logprior = 0.0;
	logprior = hier_lndmvn_jhp(alpha, alpha_prior_mean, alpha_prior_var);
		
	return (loglike + logprior);
	
}// end of statis hieroprobit_logpost

// The hieroprobitModel class stores additional data to be passed to BFGS
class hieroprobitModel {
public:
	double operator() (const Matrix<double> alpha){
/*
        const int n = y_.rows();
        const int ncat = alpha.rows() + 1;
        
        // the linear predictor
        Matrix<> mu = X_ * beta_;
        Matrix<> gamma = hier_alpha2gamma(alpha);
        
		Matrix<> cat_prob(n, ncat-1);
		//cat_prob: CATegory specific PROBability
		//the first col of cat_prob = pnorm(gamma_1 - mu)
		//thus, the col number is ncat - 1
		Matrix<> prob(n, ncat);
		for (int j=0; j< ncat-1; ++j){
			for (int i=0; i<n ; ++i){
				cat_prob(i, j) = pnorm(gamma[j+1] - mu[i], 0.0, 1.0);
			}
		}
		prob(_, ncat-1) = 1 - cat_prob(_, ncat-2);    // top category
		prob(_, 0) = cat_prob(_, 0);               // bottom category
		for (int j=1; j<(ncat-1); ++j){               // middle categories are actually differences of cumulatives
			prob(_, j) = cat_prob(_,j) - cat_prob(_, j-1);
		}
        
        // the loglikelihood
        double loglike = 0.0;
		for (int i=0; i<n; ++i){
			int ind = y_[i];
			loglike += std::log(prob(i,ind-1));
		}        
		return -1 * loglike;
*/
	}

	// Full-forms:
	Matrix<double> y_;
	Matrix<double> X_;
	Matrix<double> U_;
	Matrix<double> W_tilde_;

	// List-ified forms:
	vector< Matrix<double> > Xlist_;
	vector< Matrix<double> > Wlist_;
	vector< Matrix<double> > Ulist_;
	vector< Matrix<double> > Xilist_;

	// Parameters:
	Matrix<double> beta_;
	Matrix<double> xi_tilde_;
	Matrix<double> Sigma_xi_;
};   

//
// MCMChieroprobit implementation.  Takes Matrix<> reference which it
// fills with the posterior.
//

template <typename RNGTYPE>
void MCMChieroprobit_impl (rng<RNGTYPE>& stream, 
	const int * Y,
	const Matrix<>& nY, 
	const Matrix<>& X,
	Matrix<>& U,
	const Matrix<>& W_tilde,
	Matrix<>& xi_tilde,
	vector< Matrix<> >& Xlist,
	vector< Matrix<> >& Wlist,
	vector< Matrix<> >& Ulist,
	vector< Matrix<> >& Xilist,
	Matrix<>& beta, 
	Matrix<>& gamma,
	const Matrix<>& m,
	const Matrix<>& b0,
	const Matrix<>& B0,
	const Matrix<>& alpha_prior_mean, // a0
	const Matrix<>& alpha_prior_var,  // A0
	const double r0, 
	const Matrix<>& R0,
	const int burnin, 
	const int mcmc, 
	const int thin, 
	const int verbose, 
	const Matrix<>& tune, 
	const double tdf, 
	const int printevery, 
	Matrix<>& result) {
        
	Rprintf("Starting MCMC implementation...\n");

	// define constants and from cross-product matrices
	const unsigned int tot_iter = burnin + mcmc;  // total number of mcmc iterations
	const unsigned int nstore = mcmc / thin;      // number of draws to store
	const unsigned int k = X.cols();
	const unsigned int nclusters = Wlist.size();
	const unsigned int N = X.rows();
	const unsigned int p = Wlist[0].cols(); // dim(xi_i) 
	const int ncat = gamma.rows() - 1;

	// storage matrix or matrices
	Matrix<> storemat(nstore, k+ncat+1+(p*(p+1)/2));

	if (verbose){
		Rprintf("Categories: %d\n",ncat);
		Rprintf("Beta's: %u\n",k);
		Rprintf("Clusters: %u\n",nclusters);
		Rprintf("N: %u\n",N);
	}

	// Useful constant cross-products:
	if (verbose){
		Rprintf("Computing cross products for each of the ");
		Rprintf("%u clusters...\n",nclusters);
	}
	const Matrix<> XpX = crossprod(X);
	vector< Matrix<double> > WpWlist(nclusters);
	vector< Matrix<double> > XpXlist(nclusters);
	for (unsigned int i=0; i<nclusters; i++){
		WpWlist[i] = crossprod(Wlist[i]);
		XpXlist[i] = crossprod(Xlist[i]);
	}

/*
for (unsigned int i=0; i<nclusters; i++){
Rprintf("W[%u]:\n\n",i);
print_matrix(Wlist[i]);
Rprintf("\nW[%u]W[%u]':\n\n",i,i);
print_matrix(WpWlist[i]);
}
*/
	if (verbose)
		Rprintf("done. Computing posterior constants...\n");

	// Changing posterior components:
	Matrix<> beta_posterior_mean;
	Matrix<> Sigma_xi_posterior_scale;

	// Useful constant posterior components:
	const Matrix<> beta_posterior_precision = (XpX + B0);
	const Matrix<> beta_posterior_variance = invpd(beta_posterior_precision);
	const double Sigma_xi_posterior_df = (r0 + nclusters);

	// Prior mean of xi_{i}'s:
	const Matrix<> Zero_xi_vec(p,1);

	if (verbose)
		Rprintf("done. Initializing the linear predictor...\n");

	// initialize LinearPredictor
	Matrix<> LinearPredictor = X * beta + W_tilde * xi_tilde; //tilde's are full-versions

	if (verbose)
		Rprintf("done. Initializing the Sigma_xi components...\n");

        // initialize Sigma_xi
        Matrix<> Sigma_xi(p,p);
	for (unsigned int i=0; i<p; i++){
		for (unsigned int j=0; j<p; j++){
			if (i==j){
				Sigma_xi(i,j) = 1.0;
			} else {
				Sigma_xi(i,j) = 0.0;
			}
		}
	}
        Matrix<> Sigma_xi_Inverse = invpd(Sigma_xi);

	// Gibbs loop
	unsigned int sofar;
	int count = 0;
	int accepts = 0;

	if (verbose)
		Rprintf("done. Initializing gammas/alphas...\n");

	// Current state of the alphas:
	Matrix<> alpha = hier_gamma2alpha(gamma);

	double tune_double = tune[0];

	if (verbose)
		Rprintf("done. Initializing model structures...\n");

	// Set-up the model structure:
	hieroprobitModel hieroprobit_model;

	// Full-forms:
	hieroprobit_model.y_ = nY;
	hieroprobit_model.X_ = X;
	hieroprobit_model.U_ = U;
	hieroprobit_model.W_tilde_ = W_tilde;

	// Parameters:
	hieroprobit_model.beta_ = beta;
	hieroprobit_model.xi_tilde_ = xi_tilde;
	hieroprobit_model.Sigma_xi_ = Sigma_xi;

	// List-ified elements:
	hieroprobit_model.Xlist_  = Xlist;
	hieroprobit_model.Wlist_  = Wlist;
	hieroprobit_model.Ulist_  = Ulist;
	hieroprobit_model.Xilist_ = Xilist;

	for (unsigned int iter = 0; iter < tot_iter; ++iter) {		

		// Step 1: [U| gamma, beta, xi, y] 
		LinearPredictor = X * beta + W_tilde * xi_tilde;
		for (unsigned int i=0; i<N; i++){
			U[i] = stream.rtnorm_combo(LinearPredictor(i,0), 1.0, gamma[Y[i]-1], gamma[Y[i]]);
		}
		// Update Ulist with new U_i's:
		sofar = 0;
		for (unsigned int i=0; i<nclusters; ++i){
			for (unsigned int j=0; j<m(i,0); j++){
				Ulist[i](j,0) = U[sofar];
				sofar++;
			}
		}
		if (verbose){
			Rprintf("U after iteration %u:\n",iter);
			print_matrix(U);
		}

		// Step 2: [gamma | U, beta, xi, Sigma_xi]
		// Sample component-wise -- nrow(gamma) = ncat+1.
		// gamma[0] = -300
		// gamma[1] = 0.0
		// gamma[2-(ncat-1)] = ...
		// gamma[ncat] = 300
		for (int i=2; i<ncat; i++){
			double gamma_i_lo = gamma[i-1];
			double gamma_i_hi = gamma[i+1];
			// Compute the conditional range of gamma_i:
			for (unsigned int j=0; j<N; j++){
				// Serves as an upper bound for U's s.t. Y==(i-1)
				if (Y[j]==i){
					if (U[j]>gamma_i_lo){
						gamma_i_lo = U[j];
					}
					if (verbose){
						Rprintf("Y[%u] = %d, U[%u] = %g, gamma_lo = %g\n",j,Y[j],j,U[j],gamma_i_lo);
					}
				}
				// Serves as a lower bound for U's s.t. Y==i
				if (Y[j]==(i+1)){
					if (U[j]<gamma_i_hi){
						gamma_i_hi = U[j];
					}
					if (verbose){
						Rprintf("Y[%u] = %d, U[%u] = %g, gamma_hi = %g\n",j,Y[j],j,U[j],gamma_i_hi);
					}
				}
			}
			if (verbose){
				Rprintf("Range for gamma[%u] = (%g,%g)\n",i,gamma_i_lo,gamma_i_hi);
			}
			if (gamma_i_lo>gamma_i_hi)
				Rprintf("Crap :'(\n");
			gamma[i] = gamma_i_lo + stream.runif()*(gamma_i_hi-gamma_i_lo);
			if (verbose){
				Rprintf("New gamma[%u] = %g\n",i,gamma[i]);
			}
		}

		// Step 3: [xi| U, beta, Sigma_xi, y]
		for (unsigned int i=0; i<nclusters; ++i){
			Matrix<> XiPrecision = (WpWlist[i] + Sigma_xi_Inverse);
			Matrix<> XiVar = invpd(XiPrecision);
			Matrix<> Xibeta = Xlist[i]*beta;
			Matrix<> UiMinusXibeta = Ulist[i] - Xibeta;
			Matrix<> WipYtilde = t(Wlist[i]) * UiMinusXibeta;
			Matrix<> XiMean = XiVar * WipYtilde;
			if (verbose){
				Rprintf("Drawing xi_i from N(mu,Sigma) where:\n\n");
				Rprintf("Xibeta = \n");
				print_matrix(Xibeta);
				Rprintf("UiMinusXibeta = \n");
				print_matrix(UiMinusXibeta);
				Rprintf("WipYtilde = \n");
				print_matrix(WipYtilde);
				Rprintf("XiPrecision = \n");
				print_matrix(XiPrecision);
				Rprintf("mu = \n");
				print_matrix(XiMean);
				Rprintf("Sigma = \n");
				print_matrix(XiVar);
			}
			Xilist[i] = stream.rmvnorm(XiMean, XiVar);

		}
		// Update xi_tilde with new xi_i's:
		sofar = 0;
		for (unsigned int i=0; i<nclusters; ++i){
			for (unsigned int j=0; j<p; j++){
				xi_tilde[sofar] = Xilist[i](j,0);
				sofar++;
			}
		}
		if (verbose){
			Rprintf("xi_tilde after iteration %u:\n\n",iter);
			print_matrix(xi_tilde);
		}
		
		// Step 4: [beta| U, xi, gamma]
		// Normal-Normal Regression with Y => (U - W_tilde*xi_tilde)
		beta_posterior_mean = beta_posterior_variance*( (B0 * b0) + (t(X) * (U - W_tilde * xi_tilde)) );
		beta = stream.rmvnorm(beta_posterior_mean, beta_posterior_variance);
		
		// Step 5: [Sigma_xi| xi, y]
                // Sigma_xi ~ Inverse-Wishart(r0+n, \sum_{i=1}^{n}\xi_{i}\xi_{i}^{T})
		Sigma_xi_posterior_scale = R0;
		for (unsigned int i=0; i<nclusters; i++){
			Sigma_xi_posterior_scale += (Xilist[i])*t(Xilist[i]);
		}
 		Sigma_xi = stream.riwish(Sigma_xi_posterior_df, Sigma_xi_posterior_scale);

		// Update (Sigma_xi)^{-1}
		Sigma_xi_Inverse = invpd(Sigma_xi);

		if (verbose){
			Rprintf("\nAfter iteration %u:\n",iter);
			Rprintf("\nalpha:\n\n");
			print_matrix(alpha);
			Rprintf("\ngamma:\n\n");
			print_matrix(gamma);
			Rprintf("\nbeta:\n\n");
			print_matrix(beta);
			Rprintf("\nxi_tilde:\n\n");
			print_matrix(xi_tilde);
			Rprintf("\nSigma_xi:\n\n");
			print_matrix(Sigma_xi);
			Rprintf("\nU:\n\n");
			print_matrix(U);
		}

		// store values in matrices
		if (iter >= burnin && ((iter % thin)==0)){ 
			// The betas:
			for (unsigned int j=0; j<k; j++)
				storemat(count, j) = beta(j,0);
			// The gammas:
			for (unsigned int j=0; j<(ncat+1); j++)
				storemat(count, k+j) = gamma(j,0);
			// The p(p+1)/2 elements of Sigma_xi:
			sofar = 0;	    
			for (unsigned int i=0; i<p; i++){
				for (unsigned int j=i; j<p; j++){
					storemat(count, (k+ncat+1)+sofar) = Sigma_xi(i,j);
					sofar++;
				}
			}
			count++;
		}
		
		if ((iter+1) % printevery == 0){
			Rprintf("Finished iteration %d/%d\n",iter+1,tot_iter);
		}
		// print output to stdout
		if(verbose > 0 && iter % verbose == 0){
			Rprintf("\n\nMCMChieroprobit iteration %i of %i \n", (iter+1), tot_iter);
			Rprintf("Metropolis acceptance rate for gamma = %3.5f\n\n", 
					static_cast<double>(accepts)/static_cast<double>(iter+1));		
		}
		
		R_CheckUserInterrupt(); // allow user interrupts

    }// end of MCMC

		//Rprintf("\n\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
		//Rprintf("The Metropolis acceptance rate for beta was %3.5f", 
		//		static_cast<double>(accepts) / static_cast<double>(tot_iter));
		//Rprintf("\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");	
	
	result = storemat;

	if (verbose){
		Rprintf("Results matrix:\n");
		print_matrix(result);
	}

	return;
}


extern "C"{
  
  void MCMChieroprobit(double *sampledata, const int *samplerow, const int *samplecol, 
		const int *Y, 
		const double *nYdata, const int *nYrow, const int *nYcol, // The full Y matrix
		const double *Xdata, const int *Xrow, const int *Xcol,    // The full X matrix
		const double *Udata, const int *Urow, const int *Ucol,    // The full U column vector
		const double *mdata, const int *mrow, const int *mcol,    // Vector of cluster sizes
		const double *fullWdata, const int *fullWrow, const int *fullWcol,    // The compact W matrix
		const double *tildeWdata, const int *tildeWrow, const int *tildeWcol,    // The full Wtilde matrix
		const double *xi_tildedata, const int *xi_tilderow, const int *xi_tildecol, // The full xi vector
		const int *burnin, const int *mcmc, const int *thin, 
		const double *tunedata, const int *tunerow, const int *tunecol,
		const double* tdf,
		const int *uselecuyer, const int *seedarray, 
		const int *lecuyerstream, const int *verbose, 
		const double *betadata, const int *betarow, const int *betacol,
		const double* gammadata, const int* gammarow, const int* gammacol,
		const double *b0data, const int *b0row, const int *b0col, 
		const double *B0data, const int *B0row, const int *B0col,
		const double *a0data, const int *a0row, const int *a0col, 
		const double *A0data, const int *A0row, const int *A0col,
		const double *r0,
		const double *R0data, const int *R0row, const int *R0col,
		const int *printevery)
{  
    
	// Pull together Matrix objects:
 
	const Matrix <> nY(*nYrow, *nYcol, nYdata);
	const Matrix <> X(*Xrow, *Xcol, Xdata);
	const Matrix <> W_tilde(*tildeWrow, *tildeWcol, tildeWdata);
	const Matrix <> W(*fullWrow, *fullWcol, fullWdata);
	Matrix <> U(*Urow, *Ucol, Udata);

	// Starting states of each xi_i, concatenated into one long vector:
	Matrix <> xi_tilde(*xi_tilderow, *xi_tildecol, xi_tildedata);

	// The vector of cluster-sizes: sum(m) = N, total number of obs.
	const Matrix <> m(*mrow, *mcol, mdata);
	const int nclusters = m.rows();

	// The list-ified versions:
	if (*verbose){
		Rprintf("Creating lists...\n");
	}
	vector< Matrix<> > Xlist(nclusters);
	vector< Matrix<> > Wlist(nclusters);
	vector< Matrix<> > Ulist(nclusters);
	vector< Matrix<> > Xilist(nclusters);
	unsigned int sofar = 0;
	unsigned int xt_sofar = 0;
	// For each cluster:
	for (unsigned int i=0; i<nclusters; i++){

		// Allocate the cluster matrices:
		Matrix<> tmp_X_matrix((unsigned int)m[i],X.cols());
		Xlist[i] = tmp_X_matrix;
		Matrix<> tmp_W_matrix((unsigned int)m[i],W.cols());
		Wlist[i] = tmp_W_matrix;
		Matrix<> tmp_U_matrix((unsigned int)m[i],1);
		Ulist[i] = tmp_U_matrix;
		Matrix<> tmp_Xi_matrix(W.cols(),1);
		Xilist[i] = tmp_Xi_matrix;

		// Extract the xi_tilde elements:
		for (unsigned int k=0; k<W.cols(); k++){
			Xilist[i](k,0) = xi_tilde(xt_sofar,0); // xi_tilde is just a column vector
			xt_sofar++;
		}

		// For each observation in the cluster:
		for (unsigned int j=0; j<m[i]; j++){

			// Extract the X matrix row:
			for (unsigned int k=0; k<X.cols(); k++){
				Xlist[i](j,k) = X(sofar,k);
			}

			// Extract the W matrix row:
			for (unsigned int k=0; k<W.cols(); k++){
				Wlist[i](j,k) = W(sofar,k);
			}

			// Extract the U matrix element:
			Ulist[i](j,0) = U(sofar,0); // U is just a column vector

			// Increment the number of filled rows and move on...
			sofar++;
		}
	}
/*
	Rprintf("List Elements:\n\n");
	for (unsigned int i=0; i<nclusters; i++){
		Rprintf("List Elements[[%u]]:\n",i);
		Rprintf("X[[%u]]:\n",i);
		std::cout << Xlist[i];
		Rprintf("W[[%u]]:\n",i);
		std::cout << Wlist[i];
		Rprintf("U[[%u]]:\n",i);
		std::cout << Ulist[i];
		Rprintf("xi_tilde[[%u]]:\n",i);
		std::cout << Xilist[i];
	}
*/
	Matrix <> beta(*betarow, *betacol, betadata);
	Matrix <> gamma(*gammarow, *gammacol, gammadata);

	const Matrix <> b0(*b0row, *b0col, b0data);
	const Matrix <> B0(*B0row, *B0col, B0data);

	const Matrix <> alpha_prior_mean(*a0row, *a0col, a0data);
	const Matrix <> alpha_prior_var(*A0row, *A0col, A0data);

	const Matrix <> R0(*R0row, *R0col, R0data);

	const Matrix<> tune(*tunerow, *tunecol, tunedata);
	Matrix<> storagematrix;
	
	if (*verbose){
		Rprintf("Summary:\n");
		Rprintf("dim(nY):       %d x %d\n",nY.rows(),nY.cols());
		Rprintf("dim(X):        %d x %d\n",X.rows(),X.cols());
		Rprintf("dim(W_tilde):  %d x %d\n",W_tilde.rows(),W_tilde.cols());
		Rprintf("dim(xi_tilde): %d x %d\n",xi_tilde.rows(),xi_tilde.cols());
		Rprintf("dim(W):        %d x %d\n",W.rows(),W.cols());
		Rprintf("dim(U):        %d x %d\n",U.rows(),U.cols());
		Rprintf("dim(beta):     %d x %d\n",beta.rows(),beta.cols());
		Rprintf("dim(gamma):    %d x %d\n",gamma.rows(),gamma.cols());
		Rprintf("dim(m):        %d x %d\n",m.rows(),m.cols());
		Rprintf("Number of clusters: %d\n",nclusters);
		Rprintf("\nCalling to MCMChieroprobit_impl...\n");
	}

	MCMCPACK_PASSRNG2MODEL(MCMChieroprobit_impl, Y, nY, X, U,  // DONE 
		W_tilde, xi_tilde,                                 // DONE
		Xlist, Wlist, Ulist, Xilist,                       // DONE
		beta, gamma,                                       // DONE
		m, b0, B0,                                         // DONE
		alpha_prior_mean, alpha_prior_var,                 // DONE
		*r0, R0,                                           // DONE
		*burnin, *mcmc, *thin, *verbose, tune, *tdf,       // DONE
		*printevery,                                       // DONE
		storagematrix);                                    // DONE

	if (*verbose)
		Rprintf("Done. Returned from call to MCMChieroprobit_impl. Tidying...\n");

	const unsigned int size = *samplerow * *samplecol;
	for (unsigned int i = 0; i < size; ++i)
		sampledata[i] = storagematrix(i);

	if (*verbose)
		Rprintf("Done. Returning to R...\n");

	return;

} // End function

} // Close extern

#endif
